@@ -1,5 +1,6 @@
 from __future__ import annotations
 
+import contextlib
 import re
 import warnings
 from collections.abc import Callable, Hashable
@@ -11,7 +12,7 @@ import numpy as np
 import pandas as pd
 from pandas.errors import OutOfBoundsDatetime, OutOfBoundsTimedelta
 
-from xarray.coding.variables import (
+from xarray.coding.common import (
     SerializationWarning,
     VariableCoder,
     lazy_elemwise_func,
@@ -20,15 +21,15 @@ from xarray.coding.variables import (
     unpack_for_decoding,
     unpack_for_encoding,
 )
+from xarray.compat.pdcompat import default_precision_timestamp, timestamp_as_unit
 from xarray.core import indexing
 from xarray.core.common import contains_cftime_datetimes, is_np_datetime_like
-from xarray.core.duck_array_ops import asarray, ravel, reshape
+from xarray.core.duck_array_ops import array_all, asarray, ravel, reshape
 from xarray.core.formatting import first_n_items, format_timestamp, last_item
-from xarray.core.pdcompat import default_precision_timestamp, timestamp_as_unit
 from xarray.core.utils import attempt_import, emit_user_level_warning
 from xarray.core.variable import Variable
 from xarray.namedarray.parallelcompat import T_ChunkedArray, get_chunked_array_type
-from xarray.namedarray.pycompat import is_chunked_array
+from xarray.namedarray.pycompat import is_chunked_array, to_numpy
 from xarray.namedarray.utils import is_duck_dask_array
 
 try:
@@ -92,6 +93,12 @@ TIME_UNITS = frozenset(
 )
 
 
+_INVALID_LITERAL_TIMEDELTA64_ENCODING_KEYS = [
+    "add_offset",
+    "scale_factor",
+]
+
+
 def _is_standard_calendar(calendar: str) -> bool:
     return calendar.lower() in _STANDARD_CALENDARS
 
@@ -153,6 +160,12 @@ def _numpy_to_netcdf_timeunit(units: NPDatetimeUnitOptions) -> str:
     }[units]
 
 
+def _numpy_dtype_to_netcdf_timeunit(dtype: np.dtype) -> str:
+    unit, _ = np.datetime_data(dtype)
+    unit = cast(NPDatetimeUnitOptions, unit)
+    return _numpy_to_netcdf_timeunit(unit)
+
+
 def _ensure_padded_year(ref_date: str) -> str:
     # Reference dates without a padded year (e.g. since 1-1-1 or since 2-3-4)
     # are ambiguous (is it YMD or DMY?). This can lead to some very odd
@@ -233,7 +246,7 @@ def build_pattern(
     ]
     pattern_list = []
     for sep, name, sub_pattern in pieces:
-        pattern_list.append((sep if sep else "") + named(name, sub_pattern))
+        pattern_list.append((sep or "") + named(name, sub_pattern))
         # TODO: allow timezone offsets?
     return "^" + trailing_optional(pattern_list) + "$"
 
@@ -292,6 +305,19 @@ def _unpack_time_unit_and_ref_date(
     return time_unit, ref_date
 
 
+def _unpack_time_units_and_ref_date_cftime(units: str, calendar: str):
+    # same as _unpack_netcdf_time_units but finalizes ref_date for
+    # processing in encode_cf_datetime
+    time_units, ref_date = _unpack_netcdf_time_units(units)
+    ref_date = cftime.num2date(
+        0,
+        units=f"microseconds since {ref_date}",
+        calendar=calendar,
+        only_use_cftime_datetimes=True,
+    )
+    return time_units, ref_date
+
+
 def _decode_cf_datetime_dtype(
     data,
     units: str,
@@ -304,7 +330,7 @@ def _decode_cf_datetime_dtype(
     # Dataset.__repr__ when users try to view their lazily decoded array.
     values = indexing.ImplicitToExplicitIndexingAdapter(indexing.as_indexable(data))
     example_value = np.concatenate(
-        [first_n_items(values, 1) or [0], last_item(values) or [0]]
+        [to_numpy(first_n_items(values, 1)), to_numpy(last_item(values))]
     )
 
     try:
@@ -404,12 +430,11 @@ def _check_date_is_after_shift(
     # if we are outside the well-defined date range
     # proleptic_gregorian and standard/gregorian are only equivalent
     # if reference date and date range is >= 1582-10-15
-    if calendar != "proleptic_gregorian":
-        if date < type(date)(1582, 10, 15):
-            raise OutOfBoundsDatetime(
-                f"Dates before 1582-10-15 cannot be decoded "
-                f"with pandas using {calendar!r} calendar: {date}"
-            )
+    if calendar != "proleptic_gregorian" and date < type(date)(1582, 10, 15):
+        raise OutOfBoundsDatetime(
+            f"Dates before 1582-10-15 cannot be decoded "
+            f"with pandas using {calendar!r} calendar: {date}"
+        )
 
 
 def _check_higher_resolution(
@@ -482,7 +507,7 @@ def _decode_datetime_with_pandas(
         flat_num_dates = flat_num_dates.astype(np.float64)
 
     timedeltas = _numbers_to_timedelta(
-        flat_num_dates, time_unit, ref_date.unit, "datetime"
+        flat_num_dates, time_unit, ref_date.unit, "datetimes"
     )
 
     # add timedeltas to ref_date
@@ -510,7 +535,7 @@ def decode_cf_datetime(
     --------
     cftime.num2date
     """
-    num_dates = np.asarray(num_dates)
+    num_dates = to_numpy(num_dates)
     flat_num_dates = ravel(num_dates)
     if calendar is None:
         calendar = "standard"
@@ -554,9 +579,8 @@ def decode_cf_datetime(
                     "'time_unit' or specify 'use_cftime=True'.",
                     SerializationWarning,
                 )
-            else:
-                if _is_standard_calendar(calendar):
-                    dates = cftime_to_nptime(dates, time_unit=time_unit)
+            elif _is_standard_calendar(calendar):
+                dates = cftime_to_nptime(dates, time_unit=time_unit)
     elif use_cftime:
         dates = _decode_datetime_with_cftime(flat_num_dates, units, calendar)
     else:
@@ -576,6 +600,7 @@ def _numbers_to_timedelta(
     time_unit: NPDatetimeUnitOptions,
     ref_unit: PDDatetimeUnitOptions,
     datatype: str,
+    target_unit: PDDatetimeUnitOptions | None = None,
 ) -> np.ndarray:
     """Transform numbers to np.timedelta64."""
     # keep NaT/nan mask
@@ -599,13 +624,23 @@ def _numbers_to_timedelta(
             flat_num, cast(PDDatetimeUnitOptions, time_unit)
         )
         if time_unit != new_time_unit:
-            msg = (
-                f"Can't decode floating point {datatype} to {time_unit!r} without "
-                f"precision loss, decoding to {new_time_unit!r} instead. "
-                f"To silence this warning use time_unit={new_time_unit!r} in call to "
-                f"decoding function."
-            )
-            emit_user_level_warning(msg, SerializationWarning)
+            if target_unit is None or np.timedelta64(1, target_unit) > np.timedelta64(
+                1, new_time_unit
+            ):
+                if datatype == "datetimes":
+                    kwarg = "decode_times"
+                    coder = "CFDatetimeCoder"
+                else:
+                    kwarg = "decode_timedelta"
+                    coder = "CFTimedeltaCoder"
+                formatted_kwarg = f"{kwarg}={coder}(time_unit={new_time_unit!r})"
+                message = (
+                    f"Can't decode floating point {datatype} to {time_unit!r} "
+                    f"without precision loss; decoding to {new_time_unit!r} "
+                    f"instead. To silence this warning pass {formatted_kwarg} "
+                    f"to your opening function."
+                )
+                emit_user_level_warning(message, SerializationWarning)
             time_unit = new_time_unit
 
     # Cast input ordinals to integers and properly handle NaN/NaT
@@ -626,23 +661,31 @@ def decode_cf_timedelta(
     """Given an array of numeric timedeltas in netCDF format, convert it into a
     numpy timedelta64 ["s", "ms", "us", "ns"] array.
     """
-    num_timedeltas = np.asarray(num_timedeltas)
+    num_timedeltas = to_numpy(num_timedeltas)
     unit = _netcdf_to_numpy_timeunit(units)
 
+    # special case empty arrays
+    is_empty_array = num_timedeltas.size == 0
+
     with warnings.catch_warnings():
         warnings.filterwarnings("ignore", "All-NaN slice encountered", RuntimeWarning)
-        _check_timedelta_range(np.nanmin(num_timedeltas), unit, time_unit)
-        _check_timedelta_range(np.nanmax(num_timedeltas), unit, time_unit)
+        if not is_empty_array:
+            _check_timedelta_range(np.nanmin(num_timedeltas), unit, time_unit)
+            _check_timedelta_range(np.nanmax(num_timedeltas), unit, time_unit)
 
-    timedeltas = _numbers_to_timedelta(num_timedeltas, unit, "s", "timedelta")
+    timedeltas = _numbers_to_timedelta(
+        num_timedeltas, unit, "s", "timedeltas", target_unit=time_unit
+    )
     pd_timedeltas = pd.to_timedelta(ravel(timedeltas))
 
-    if np.isnat(timedeltas).all():
+    if not is_empty_array and np.isnat(timedeltas).all():
         empirical_unit = time_unit
     else:
         empirical_unit = pd_timedeltas.unit
 
-    if np.timedelta64(1, time_unit) > np.timedelta64(1, empirical_unit):
+    if is_empty_array or np.timedelta64(1, time_unit) > np.timedelta64(
+        1, empirical_unit
+    ):
         time_unit = empirical_unit
 
     if time_unit not in {"s", "ms", "us", "ns"}:
@@ -667,6 +710,7 @@ def _infer_time_units_from_diff(unique_timedeltas) -> str:
     # todo: check, if this function works correctly wrt np.timedelta64
     unit_timedelta: Callable[[str], timedelta] | Callable[[str], np.timedelta64]
     zero_timedelta: timedelta | np.timedelta64
+    unique_timedeltas = asarray(unique_timedeltas)
     if unique_timedeltas.dtype == np.dtype("O"):
         time_units = _NETCDF_TIME_UNITS_CFTIME
         unit_timedelta = _unit_timedelta_cftime
@@ -676,11 +720,15 @@ def _infer_time_units_from_diff(unique_timedeltas) -> str:
         unit_timedelta = _unit_timedelta_numpy
         zero_timedelta = np.timedelta64(0, "ns")
     for time_unit in time_units:
-        if np.all(unique_timedeltas % unit_timedelta(time_unit) == zero_timedelta):
+        if array_all(unique_timedeltas % unit_timedelta(time_unit) == zero_timedelta):
             return time_unit
     return "seconds"
 
 
+def _time_units_to_timedelta(units: str) -> timedelta:
+    return timedelta(microseconds=_US_PER_TIME_DELTA[units])
+
+
 def infer_calendar_name(dates) -> CFCalendar:
     """Given an array of datetimes, infer the CF calendar name"""
     if is_np_datetime_like(dates.dtype):
@@ -880,12 +928,10 @@ def _cleanup_netcdf_time_units(units: str) -> str:
     time_units = time_units.lower()
     if not time_units.endswith("s"):
         time_units = f"{time_units}s"
-    try:
+    # don't worry about reifying the units if they're out of bounds or
+    # formatted badly
+    with contextlib.suppress(OutOfBoundsDatetime, ValueError):
         units = f"{time_units} since {format_timestamp(ref_date)}"
-    except (OutOfBoundsDatetime, ValueError):
-        # don't worry about reifying the units if they're out of bounds or
-        # formatted badly
-        pass
     return units
 
 
@@ -900,46 +946,43 @@ def _encode_datetime_with_cftime(dates, units: str, calendar: str) -> np.ndarray
     else:
         cftime = attempt_import("cftime")
 
+    dates = np.asarray(dates)
+    original_shape = dates.shape
+
     if np.issubdtype(dates.dtype, np.datetime64):
         # numpy's broken datetime conversion only works for us precision
         dates = dates.astype("M8[us]").astype(datetime)
 
-    def wrap_dt(dt):
-        # convert to cftime proleptic gregorian in case of datetime.datetime
-        # needed because of https://github.com/Unidata/cftime/issues/354
-        if isinstance(dt, datetime) and not isinstance(dt, cftime.datetime):
-            dt = cftime.datetime(
-                dt.year,
-                dt.month,
-                dt.day,
-                dt.hour,
-                dt.minute,
-                dt.second,
-                dt.microsecond,
-                calendar="proleptic_gregorian",
-            )
-        return dt
+    dates = np.atleast_1d(dates)
 
-    def encode_datetime(d):
-        # Since netCDF files do not support storing float128 values, we ensure
-        # that float64 values are used by setting longdouble=False in num2date.
-        # This try except logic can be removed when xarray's minimum version of
-        # cftime is at least 1.6.2.
-        try:
-            return (
-                np.nan
-                if d is None
-                else cftime.date2num(wrap_dt(d), units, calendar, longdouble=False)
-            )
-        except TypeError:
-            return np.nan if d is None else cftime.date2num(wrap_dt(d), units, calendar)
+    # Find all the None position
+    none_position = dates == None  # noqa: E711
+    filtered_dates = dates[~none_position]
 
-    return reshape(np.array([encode_datetime(d) for d in ravel(dates)]), dates.shape)
+    # Since netCDF files do not support storing float128 values, we ensure
+    # that float64 values are used by setting longdouble=False in num2date.
+    # This try except logic can be removed when xarray's minimum version of
+    # cftime is at least 1.6.2.
+    try:
+        encoded_nums = cftime.date2num(
+            filtered_dates, units, calendar, longdouble=False
+        )
+    except TypeError:
+        encoded_nums = cftime.date2num(filtered_dates, units, calendar)
+
+    if filtered_dates.size == none_position.size:
+        return encoded_nums.reshape(original_shape)
+
+    # Create a full matrix of NaN
+    # And fill the num dates in the not NaN or None position
+    result = np.full(dates.shape, np.nan)
+    result[np.nonzero(~none_position)] = encoded_nums
+    return result.reshape(original_shape)
 
 
 def cast_to_int_if_safe(num) -> np.ndarray:
     int_num = np.asarray(num, dtype=np.int64)
-    if (num == int_num).all():
+    if array_all(num == int_num):
         num = int_num
     return num
 
@@ -955,42 +998,6 @@ def _division(deltas, delta, floor):
     return num
 
 
-def _cast_to_dtype_if_safe(num: np.ndarray, dtype: np.dtype) -> np.ndarray:
-    with warnings.catch_warnings():
-        warnings.filterwarnings("ignore", message="overflow")
-        cast_num = np.asarray(num, dtype=dtype)
-
-    if np.issubdtype(dtype, np.integer):
-        if not (num == cast_num).all():
-            if np.issubdtype(num.dtype, np.floating):
-                raise ValueError(
-                    f"Not possible to cast all encoded times from "
-                    f"{num.dtype!r} to {dtype!r} without losing precision. "
-                    f"Consider modifying the units such that integer values "
-                    f"can be used, or removing the units and dtype encoding, "
-                    f"at which point xarray will make an appropriate choice."
-                )
-            else:
-                raise OverflowError(
-                    f"Not possible to cast encoded times from "
-                    f"{num.dtype!r} to {dtype!r} without overflow. Consider "
-                    f"removing the dtype encoding, at which point xarray will "
-                    f"make an appropriate choice, or explicitly switching to "
-                    "a larger integer dtype."
-                )
-    else:
-        if np.isinf(cast_num).any():
-            raise OverflowError(
-                f"Not possible to cast encoded times from {num.dtype!r} to "
-                f"{dtype!r} without overflow.  Consider removing the dtype "
-                f"encoding, at which point xarray will make an appropriate "
-                f"choice, or explicitly switching to a larger floating point "
-                f"dtype."
-            )
-
-    return cast_num
-
-
 def encode_cf_datetime(
     dates: T_DuckArray,  # type: ignore[misc]
     units: str | None = None,
@@ -1013,6 +1020,27 @@ def encode_cf_datetime(
         return _eagerly_encode_cf_datetime(dates, units, calendar, dtype)
 
 
+def _infer_needed_units_numpy(ref_date, data_units):
+    needed_units, data_ref_date = _unpack_time_unit_and_ref_date(data_units)
+    needed_units = _numpy_to_netcdf_timeunit(needed_units)
+    ref_delta = abs(data_ref_date - ref_date).to_timedelta64()
+    data_delta = _unit_timedelta_numpy(needed_units)
+    if (ref_delta % data_delta) > np.timedelta64(0, "ns"):
+        needed_units = _infer_time_units_from_diff(ref_delta)
+    return needed_units
+
+
+def _infer_needed_units_cftime(ref_date, data_units, calendar):
+    needed_units, data_ref_date = _unpack_time_units_and_ref_date_cftime(
+        data_units, calendar
+    )
+    ref_delta = abs(data_ref_date - ref_date)
+    data_delta = _time_units_to_timedelta(needed_units)
+    if (ref_delta % data_delta) > timedelta(seconds=0):
+        needed_units = _infer_time_units_from_diff(ref_delta)
+    return needed_units
+
+
 def _eagerly_encode_cf_datetime(
     dates: T_DuckArray,  # type: ignore[misc]
     units: str | None = None,
@@ -1030,6 +1058,8 @@ def _eagerly_encode_cf_datetime(
     if calendar is None:
         calendar = infer_calendar_name(dates)
 
+    raise_incompatible_units_error = False
+    raise_gregorian_proleptic_gregorian_mismatch_error = False
     try:
         if not _is_standard_calendar(calendar) or dates.dtype.kind == "O":
             # parse with cftime instead
@@ -1038,16 +1068,7 @@ def _eagerly_encode_cf_datetime(
         if calendar in ["standard", "gregorian"] and np.nanmin(dates).astype(
             "=M8[us]"
         ).astype(datetime) < datetime(1582, 10, 15):
-            # if we use standard calendar and for dates before the reform
-            # we need to use cftime instead
-            emit_user_level_warning(
-                f"Unable to encode numpy.datetime64 objects with {calendar} calendar."
-                "Using cftime.datetime objects instead, reason: dates prior "
-                "reform date (1582-10-15). To silence this warning transform "
-                "numpy.datetime64 to corresponding cftime.datetime beforehand.",
-                SerializationWarning,
-            )
-            raise OutOfBoundsDatetime
+            raise_gregorian_proleptic_gregorian_mismatch_error = True
 
         time_unit, ref_date = _unpack_time_unit_and_ref_date(units)
         # calendar equivalence only for days after the reform
@@ -1062,16 +1083,7 @@ def _eagerly_encode_cf_datetime(
         time_deltas = dates_as_index - ref_date
 
         # retrieve needed units to faithfully encode to int64
-        needed_unit, data_ref_date = _unpack_time_unit_and_ref_date(data_units)
-        needed_units = _numpy_to_netcdf_timeunit(needed_unit)
-        if data_units != units:
-            # this accounts for differences in the reference times
-            ref_delta = abs(data_ref_date - ref_date).to_timedelta64()
-            data_delta = np.timedelta64(1, needed_unit)
-            if (ref_delta % data_delta) > np.timedelta64(0, "ns"):
-                needed_units = _infer_time_units_from_diff(ref_delta)
-
-        # needed time delta to encode faithfully to int64
+        needed_units = _infer_needed_units_numpy(ref_date, data_units)
         needed_time_delta = _unit_timedelta_numpy(needed_units)
 
         floor_division = np.issubdtype(dtype, np.integer) or dtype is None
@@ -1085,7 +1097,6 @@ def _eagerly_encode_cf_datetime(
                     f"Set encoding['dtype'] to floating point dtype to silence this warning."
                 )
             elif np.issubdtype(dtype, np.integer) and allow_units_modification:
-                floor_division = True
                 new_units = f"{needed_units} since {format_timestamp(ref_date)}"
                 emit_user_level_warning(
                     f"Times can't be serialized faithfully to int64 with requested units {units!r}. "
@@ -1095,6 +1106,10 @@ def _eagerly_encode_cf_datetime(
                 )
                 units = new_units
                 time_delta = needed_time_delta
+                floor_division = True
+            elif np.issubdtype(dtype, np.integer) and not allow_units_modification:
+                new_units = f"{needed_units} since {format_timestamp(ref_date)}"
+                raise_incompatible_units_error = True
 
         # get resolution of TimedeltaIndex and align time_delta
         # todo: check, if this works in any case
@@ -1104,14 +1119,49 @@ def _eagerly_encode_cf_datetime(
         num = reshape(num.values, dates.shape)
 
     except (OutOfBoundsDatetime, OverflowError, ValueError):
+        time_units, ref_date = _unpack_time_units_and_ref_date_cftime(units, calendar)
+        time_delta_cftime = _time_units_to_timedelta(time_units)
+        needed_units = _infer_needed_units_cftime(ref_date, data_units, calendar)
+        needed_time_delta_cftime = _time_units_to_timedelta(needed_units)
+
+        if (
+            np.issubdtype(dtype, np.integer)
+            and time_delta_cftime > needed_time_delta_cftime
+        ):
+            new_units = f"{needed_units} since {format_cftime_datetime(ref_date)}"
+            if allow_units_modification:
+                emit_user_level_warning(
+                    f"Times can't be serialized faithfully to int64 with requested units {units!r}. "
+                    f"Serializing with units {new_units!r} instead. "
+                    f"Set encoding['dtype'] to floating point dtype to serialize with units {units!r}. "
+                    f"Set encoding['units'] to {new_units!r} to silence this warning ."
+                )
+                units = new_units
+            else:
+                raise_incompatible_units_error = True
+
         num = _encode_datetime_with_cftime(dates, units, calendar)
         # do it now only for cftime-based flow
         # we already covered for this in pandas-based flow
         num = cast_to_int_if_safe(num)
 
-    if dtype is not None:
-        # todo: check, if this is really needed for all dtypes
-        num = _cast_to_dtype_if_safe(num, dtype)
+    if raise_incompatible_units_error:
+        raise ValueError(
+            f"Times can't be serialized faithfully to int64 with requested units {units!r}. "
+            f"Consider setting encoding['dtype'] to a floating point dtype to serialize with "
+            f"units {units!r}. Consider setting encoding['units'] to {new_units!r} to "
+            f"serialize with an integer dtype."
+        )
+    if raise_gregorian_proleptic_gregorian_mismatch_error:
+        raise ValueError(
+            f"Unable to encode np.datetime64 values with {calendar} "
+            f"calendar, because some or all values are prior to the reform "
+            f"date of 1582-10-15. To encode these times, set "
+            f"encoding['calendar'] to 'proleptic_gregorian' instead, which "
+            f"is the true calendar that np.datetime64 values use. The "
+            f"'standard' or 'gregorian' calendar is only equivalent to the "
+            f"'proleptic_gregorian' calendar after the reform date."
+        )
 
     return num, units, calendar
 
@@ -1143,7 +1193,8 @@ def _lazily_encode_cf_datetime(
             units = "microseconds since 1970-01-01"
             dtype = np.dtype("int64")
         else:
-            units = "nanoseconds since 1970-01-01"
+            netcdf_unit = _numpy_dtype_to_netcdf_timeunit(dates.dtype)
+            units = f"{netcdf_unit} since 1970-01-01"
             dtype = np.dtype("int64")
 
     if units is None or dtype is None:
@@ -1187,6 +1238,9 @@ def _eagerly_encode_cf_timedelta(
     data_units = infer_timedelta_units(timedeltas)
     if units is None:
         units = data_units
+    # units take precedence in the case of zero-size array
+    if timedeltas.size == 0:
+        data_units = units
 
     time_delta = _unit_timedelta_numpy(units)
     time_deltas = pd.TimedeltaIndex(ravel(timedeltas))
@@ -1223,14 +1277,17 @@ def _eagerly_encode_cf_timedelta(
             time_delta = needed_time_delta
             time_delta = time_delta.astype(f"=m8[{deltas_unit}]")
             floor_division = True
+        elif np.issubdtype(dtype, np.integer) and not allow_units_modification:
+            raise ValueError(
+                f"Timedeltas can't be serialized faithfully to int64 with requested units {units!r}. "
+                f"Consider setting encoding['dtype'] to a floating point dtype to serialize with "
+                f"units {units!r}. Consider setting encoding['units'] to {needed_units!r} to "
+                f"serialize with an integer dtype."
+            )
 
     num = _division(time_deltas, time_delta, floor_division)
     num = reshape(num.values, timedeltas.shape)
 
-    if dtype is not None:
-        # todo: check, if this is needed for all dtypes
-        num = _cast_to_dtype_if_safe(num, dtype)
-
     return num, units
 
 
@@ -1249,7 +1306,7 @@ def _lazily_encode_cf_timedelta(
     timedeltas: T_ChunkedArray, units: str | None = None, dtype: np.dtype | None = None
 ) -> tuple[T_ChunkedArray, str]:
     if units is None and dtype is None:
-        units = "nanoseconds"
+        units = _numpy_dtype_to_netcdf_timeunit(timedeltas.dtype)
         dtype = np.dtype("int64")
 
     if units is None or dtype is None:
@@ -1309,6 +1366,12 @@ class CFDatetimeCoder(VariableCoder):
             units = encoding.pop("units", None)
             calendar = encoding.pop("calendar", None)
             dtype = encoding.get("dtype", None)
+
+            # in the case of packed data we need to encode into
+            # float first, the correct dtype will be established
+            # via CFScaleOffsetCoder/CFMaskCoder
+            if "add_offset" in encoding or "scale_factor" in encoding:
+                dtype = data.dtype if data.dtype.kind == "f" else "float64"
             (data, units, calendar) = encode_cf_datetime(data, units, calendar, dtype)
 
             safe_setitem(attrs, "units", units, name=name)
@@ -1342,55 +1405,149 @@ class CFDatetimeCoder(VariableCoder):
             return variable
 
 
+def has_timedelta64_encoding_dtype(attrs_or_encoding: dict) -> bool:
+    dtype = attrs_or_encoding.get("dtype")
+    return isinstance(dtype, str) and dtype.startswith("timedelta64")
+
+
+def resolve_time_unit_from_attrs_dtype(
+    attrs_dtype: str, name: T_Name
+) -> PDDatetimeUnitOptions:
+    dtype = np.dtype(attrs_dtype)
+    resolution, _ = np.datetime_data(dtype)
+    resolution = cast(NPDatetimeUnitOptions, resolution)
+    if np.timedelta64(1, resolution) > np.timedelta64(1, "s"):
+        time_unit = cast(PDDatetimeUnitOptions, "s")
+        message = (
+            f"Following pandas, xarray only supports decoding to timedelta64 "
+            f"values with a resolution of 's', 'ms', 'us', or 'ns'. Encoded "
+            f"values for variable {name!r} have a resolution of "
+            f"{resolution!r}. Attempting to decode to a resolution of 's'. "
+            f"Note, depending on the encoded values, this may lead to an "
+            f"OverflowError. Additionally, data will not be identically round "
+            f"tripped; xarray will choose an encoding dtype of "
+            f"'timedelta64[s]' when re-encoding."
+        )
+        emit_user_level_warning(message)
+    elif np.timedelta64(1, resolution) < np.timedelta64(1, "ns"):
+        time_unit = cast(PDDatetimeUnitOptions, "ns")
+        message = (
+            f"Following pandas, xarray only supports decoding to timedelta64 "
+            f"values with a resolution of 's', 'ms', 'us', or 'ns'. Encoded "
+            f"values for variable {name!r} have a resolution of "
+            f"{resolution!r}. Attempting to decode to a resolution of 'ns'. "
+            f"Note, depending on the encoded values, this may lead to loss of "
+            f"precision. Additionally, data will not be identically round "
+            f"tripped; xarray will choose an encoding dtype of "
+            f"'timedelta64[ns]' when re-encoding."
+        )
+        emit_user_level_warning(message)
+    else:
+        time_unit = cast(PDDatetimeUnitOptions, resolution)
+    return time_unit
+
+
 class CFTimedeltaCoder(VariableCoder):
     """Coder for CF Timedelta coding.
 
     Parameters
     ----------
     time_unit : PDDatetimeUnitOptions
-          Target resolution when decoding timedeltas. Defaults to "ns".
+        Target resolution when decoding timedeltas via units. Defaults to "ns".
+        When decoding via dtype, the resolution is specified in the dtype
+        attribute, so this parameter is ignored.
+    decode_via_units : bool
+        Whether to decode timedeltas based on the presence of a timedelta-like
+        units attribute, e.g. "seconds". Defaults to True, but in the future
+        will default to False.
+    decode_via_dtype : bool
+        Whether to decode timedeltas based on the presence of a np.timedelta64
+        dtype attribute, e.g. "timedelta64[s]". Defaults to True.
     """
 
     def __init__(
         self,
-        time_unit: PDDatetimeUnitOptions = "ns",
+        time_unit: PDDatetimeUnitOptions | None = None,
+        decode_via_units: bool = True,
+        decode_via_dtype: bool = True,
     ) -> None:
         self.time_unit = time_unit
+        self.decode_via_units = decode_via_units
+        self.decode_via_dtype = decode_via_dtype
         self._emit_decode_timedelta_future_warning = False
 
     def encode(self, variable: Variable, name: T_Name = None) -> Variable:
         if np.issubdtype(variable.data.dtype, np.timedelta64):
             dims, data, attrs, encoding = unpack_for_encoding(variable)
+            dtype = encoding.get("dtype", None)
+            units = encoding.pop("units", None)
 
-            data, units = encode_cf_timedelta(
-                data, encoding.pop("units", None), encoding.get("dtype", None)
-            )
+            # in the case of packed data we need to encode into
+            # float first, the correct dtype will be established
+            # via CFScaleOffsetCoder/CFMaskCoder
+            if "add_offset" in encoding or "scale_factor" in encoding:
+                dtype = data.dtype if data.dtype.kind == "f" else "float64"
+
+            resolution, _ = np.datetime_data(variable.dtype)
+            attrs_dtype = f"timedelta64[{resolution}]"
+            safe_setitem(attrs, "dtype", attrs_dtype, name=name)
+
+            data, units = encode_cf_timedelta(data, units, dtype)
             safe_setitem(attrs, "units", units, name=name)
-
             return Variable(dims, data, attrs, encoding, fastpath=True)
         else:
             return variable
 
     def decode(self, variable: Variable, name: T_Name = None) -> Variable:
         units = variable.attrs.get("units", None)
-        if isinstance(units, str) and units in TIME_UNITS:
-            if self._emit_decode_timedelta_future_warning:
-                emit_user_level_warning(
-                    "In a future version of xarray decode_timedelta will "
-                    "default to False rather than None. To silence this "
-                    "warning, set decode_timedelta to True, False, or a "
-                    "'CFTimedeltaCoder' instance.",
-                    FutureWarning,
-                )
+        has_timedelta_units = isinstance(units, str) and units in TIME_UNITS
+        has_timedelta_dtype = has_timedelta64_encoding_dtype(variable.attrs)
+        is_dtype_decodable = has_timedelta_units and has_timedelta_dtype
+        is_units_decodable = has_timedelta_units
+        if (is_dtype_decodable and self.decode_via_dtype) or (
+            is_units_decodable and self.decode_via_units
+        ):
             dims, data, attrs, encoding = unpack_for_decoding(variable)
-
             units = pop_to(attrs, encoding, "units")
-            dtype = np.dtype(f"timedelta64[{self.time_unit}]")
-            transform = partial(
-                decode_cf_timedelta, units=units, time_unit=self.time_unit
-            )
-            data = lazy_elemwise_func(data, transform, dtype=dtype)
+            if is_dtype_decodable:
+                attrs_dtype = attrs.pop("dtype")
+                if self.time_unit is None:
+                    time_unit = resolve_time_unit_from_attrs_dtype(attrs_dtype, name)
+                else:
+                    time_unit = self.time_unit
+            else:
+                if self._emit_decode_timedelta_future_warning:
+                    var_string = f"the variable {name!r}" if name else ""
+                    emit_user_level_warning(
+                        "In a future version, xarray will not decode "
+                        f"{var_string} into a timedelta64 dtype based on the "
+                        "presence of a timedelta-like 'units' attribute by "
+                        "default. Instead it will rely on the presence of a "
+                        "timedelta64 'dtype' attribute, which is now xarray's "
+                        "default way of encoding timedelta64 values.\n"
+                        "To continue decoding into a timedelta64 dtype, either "
+                        "set `decode_timedelta=True` when opening this "
+                        "dataset, or add the attribute "
+                        "`dtype='timedelta64[ns]'` to this variable on disk.\n"
+                        "To opt-in to future behavior, set "
+                        "`decode_timedelta=False`.",
+                        FutureWarning,
+                    )
+                if self.time_unit is None:
+                    time_unit = cast(PDDatetimeUnitOptions, "ns")
+                else:
+                    time_unit = self.time_unit
 
+                # Handle edge case that decode_via_dtype=False and
+                # decode_via_units=True, and timedeltas were encoded with a
+                # dtype attribute. We need to remove the dtype attribute
+                # to prevent an error during round tripping.
+                if has_timedelta_dtype:
+                    attrs.pop("dtype")
+
+            dtype = np.dtype(f"timedelta64[{time_unit}]")
+            transform = partial(decode_cf_timedelta, units=units, time_unit=time_unit)
+            data = lazy_elemwise_func(data, transform, dtype=dtype)
             return Variable(dims, data, attrs, encoding, fastpath=True)
         else:
             return variable
